# -*- coding: utf-8 -*-
"""
Ad-hoc secret generation command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
from __future__ import print_function

import sys

from clik import args, parser

from safe.app import ignore_file_argument
from safe.cmd.gen import gen
from safe.ec import VALIDATION_ERROR
from safe.sgen import generate


DEFAULT_COUNT = 1
DEFAULT_LENGTH = 32


@gen.bare
def gen_ad_hoc():
    """Generate and print random strings to stdout."""
    ignore_file_argument()

    generator_choices = sorted(generate)
    generator_help = 'generator to use (choices: %s) (default: ' \
                     '%%(default)s)' % ', '.join(generator_choices)
    parser.add_argument(
        '-g',
        '--generator',
        choices=generator_choices,
        default='default',
        help=generator_help,
        metavar='GENERATOR',
    )
    parser.add_argument(
        '-l',
        '--length',
        default=DEFAULT_LENGTH,
        help='length of secret to generate (default: %(default)s)',
        type=int,
    )
    parser.add_argument(
        '-c',
        '--count',
        default=DEFAULT_COUNT,
        help='number of secrets to generate (one per line) (default: '
             '%(default)s)',
        type=int,
    )

    yield

    if args.count < 1:
        print('error: -c/--count must be one or greater', file=sys.stderr)
        yield VALIDATION_ERROR

    if args.length < 1:
        print('error: -l/--length must be one or greater', file=sys.stderr)
        yield VALIDATION_ERROR

    for _ in range(args.count):
        print(generate[args.generator](args.length))



# -*- coding: utf-8 -*-
"""
Clipboard drivers.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
import os
import subprocess

from clik.compat import iteritems
from clik.util import AttributeDict

from safe.util import get_executable, Subprocess, temporary_directory


class ClipboardError(Exception):
    """Raised for errors occurring in this module."""


def sorted_by_precedence(values):
    """
    Return ``values`` sorted ascending by their ``precedence`` property.

    :param values: Sequence of values to sort (each value must have a
                   ``precedence`` property)
    :return: Sorted list of values
    :rtype: :class:`list`
    """
    return sorted(values, key=lambda value: -value.precedence)


class Registry(AttributeDict):
    """Registry for clipboard drivers."""

    def register(self, cls):
        """
        Register a clipboard driver.

        :param safe.clip.Driver cls: Driver class
        :raises: :exc:`ClipboardError` if driver with class' ``name`` has
                 already been registered
        :return: Class that was passed in
        """
        if cls.name in self:
            fmt = 'Clipboard driver "%s" already registered'
            raise ClipboardError(fmt % cls.name)
        self[cls.name] = cls
        return cls

    @property
    def supported(self):
        """
        List of supported drivers, sorted by precedence.

        :type: :class:`list` of :class:`safe.clip.Driver` classes
        """
        drivers = [cls for cls in self.values() if cls.supported]
        return sorted_by_precedence(drivers)

    @property
    def preferred(self):
        """
        Preferred clipboard driver for the system.

        :type: :class:`safe.clip.Driver`
        """
        return sorted_by_precedence(self.supported)[0]

    def configure_parser(self, parser):
        """
        Configure argument parser with clipboard arguments.

        The arguments allow the end user to select how clipboard operations
        are done.

        If there are multiple supported drivers, the ``--driver`` argument
        is added to allow the user to select the driver to use. If there is
        only one driver, the ``--driver`` argument is omitted and only the
        driver-specific arguments are configured.

        :param argparse.ArgumentParser parser: Parser to configure
        :raises: :exc:`ClipboardError` if there are no supported clipboards
        """
        if not self:
            raise ClipboardError('no supported clipboards')

        choices = sorted([cls.name for cls in self.supported])
        if len(choices) > 1:
            choices_str = ', '.join(choices)
            parser.add_argument(
                '-d',
                '--driver',
                choices=choices,
                default=self.preferred.name,
                help='clipboard driver to use (choices: %s) (default: '
                     '%%(default)s)' % choices_str,
            )

        for cls in self.supported:
            cls.configure_parser(parser)

    def driver_for_args(self, args):
        """
        Return clipboard driver for given arguments.

        :param argparse.Namespace args: Arguments supplied by end user
        :return: Clipboard driver configured per arguments
        :rtype: :class:`safe.clip.Driver`
        """
        if len(self.supported) > 1:
            cls = self[args.driver]
        else:
            cls = self[list(self.keys())[0]]
        params = {}
        for name, kwargs in iteritems(cls.parameters):
            args_name = '%s_%s' % (cls.name, name)
            value = getattr(args, args_name, None)
            if value is not None:
                params[name] = value
        return cls(**params)


#: Global registry for clipboard drivers.
#:
#: :type: :class:`Registry`
clipboard_drivers = Registry()


def run(command, stdin=None):
    """
    Run a command, send it ``stdin``, return exit code, stdout, and stderr.

    :param command: String or sequence of strings representing the command
    :param stdin: Optional value to send to stdin
    :type stdin: :class:`str` or ``None``
    :return: 3-tuple ``(returncode, stdout, stderr)``
    """
    with temporary_directory() as tmp:
        stdout_path = os.path.join(tmp, 'stdout')
        stderr_path = os.path.join(tmp, 'stderr')
        with open(stdout_path, 'w') as stdout_f, \
             open(stderr_path, 'w') as stderr_f:  # noqa: E127
            process = Subprocess(
                command,
                stdin=subprocess.PIPE,
                stdout=stdout_f,
                stderr=stderr_f,
            )
            process.communicate(stdin)
        with open(stdout_path) as f:
            stdout = f.read()
        with open(stderr_path) as f:
            stderr = f.read()
        return process.returncode, stdout, stderr


class Driver(object):
    """Driver base class."""

    #: Human-friendly name of the driver.
    #:
    #: :type: :class:`str`
    name = None

    #: Precedence of the driver. This is used to figure out the default
    #: driver if multiple drivers are supported on a system. The
    #: driver with the highest precedence is considered the preferred
    #: driver.
    #:
    #: :type: :class:`int`
    precedence = 0

    #: Parameters for the driver. This is used to expose configuration
    #: options to end users as well as define defaults.
    #:
    #: The parameters dict maps ``name -> dict()``, where the dict is
    #: the configuration for the parameter. Configs can have
    #: ``description`` and ``default`` keys, which set the help
    #: message and default value, respectively.
    #:
    #: When the driver is instantiated, the parameters are available
    #: in ``self.param``.
    #:
    #: See :class:`Pasteboard` or :class:`Xclip` for examples.
    #:
    #: :type: :class:`dict`
    parameters = None

    #: Indicates whether driver is supported.
    #:
    #: :type: :class:`bool`
    supported = False

    #: (instance) Parameter values for this instance of the driver.
    #: Params are composed of the default values, overridden by anything
    #: passed into the constructor.
    #:
    #: :type: :class:`clik.AttributeDict`
    param = None

    @classmethod
    def configure_parser(cls, parser):
        """
        Configure parameter arguments on ``parser``.

        :param argparse.ArgumentParser parser: Parser to configure
        """
        for name, kwargs in iteritems(cls.parameters):
            kwargs.setdefault('metavar', name.upper())
            name = name.replace('_', '-')
            parser.add_argument('--%s-%s' % (cls.name, name), **kwargs)

    def __init__(self, **kwargs):
        """
        Instantiate the driver.

        :param kwargs: Parameters (overrides defaults set in
                       :attr:`parameters`)
        """
        self.param = AttributeDict()
        if self.parameters:
            defaults = {}
            for name, param_kwargs in iteritems(self.parameters):
                if 'default' in param_kwargs:
                    defaults[name] = param_kwargs['default']
            self.param.update(defaults)
        self.param.update(kwargs)

    def get(self):
        """
        Stub method for getting data from the clipboard.

        :raises: :exc:`NotImplementedError`
        """
        raise NotImplementedError

    def put(self, value):
        """
        Stub method for putting data on the clipboard.

        :param str value: Value to put on the clipboard
        :raises: :exc:`NotImplementedError`
        """
        raise NotImplementedError


class Pasteboard(Driver):
    """Clipboard driver using pbcopy and pbpaste, found on macOS systems."""

    #: Absolute path to the ``pbcopy`` executable, or ``None`` if not
    #: found on the system.
    #:
    #: :type: :class:`str` or ``None``
    pbcopy = get_executable('pbcopy')

    #: Absolute path to the ``pbpaste`` executable, or ``None`` if not
    #: found on the system.
    #:
    #: :type: :class:`str` or ``None``
    pbpaste = get_executable('pbpaste')

    #: This driver is supported if pbcopy and pbpaste are present.
    supported = pbcopy is not None and pbpaste is not None

    name = 'pb'
    precedence = 1000
    parameters = dict(
        board=dict(
            default='general',
            help='clipboard to use (default: %(default)s)',
        ),
    )

    def get(self):
        """Return data on the pasteboard."""
        rc, stdout, stderr = run((self.pbpaste, '-pboard', self.param.board))
        if rc:
            fmt = '%s failed with stderr: %s'
            raise ClipboardError(fmt % (self.pbpaste, stderr))
        return stdout

    def put(self, value):
        """
        Put data on the pasteboard.

        :param str value: Value to put on the pasteboard
        """
        command = (self.pbcopy, '-pboard', self.param.board)
        rc, _, stderr = run(command, stdin=value)
        if rc:
            fmt = '%s failed with stderr: %s'
            raise ClipboardError(fmt % (self.pbcopy, stderr))


class Xclip(Driver):
    """Clipboard driver using xclip from the X Window System."""

    #: Absolute path to the ``xclip`` executable, or ``None`` if not
    #: found on the system.
    #:
    #: :type: :class:`str` or ``None``
    xclip = get_executable('xclip')

    #: This driver is supported if the xclip executable is present.
    supported = xclip is not None

    name = 'xclip'
    parameters = dict(
        selection=dict(
            default='primary',
            help='X selection to use (default: %(default)s)',
        ),
    )

    def get(self):
        """Return data on the clipboard."""
        command = (self.xclip, '-out', '-selection', self.param.selection)
        rc, stdout, stderr = run(command)
        if rc:
            fmt = '%s failed with stderr: %s'
            raise ClipboardError(fmt % (self.xclip, stderr))
        return stdout

    def put(self, value):
        """
        Put data on the clipboard.

        :param str value: Value to put on the clipboard
        """
        command = (self.xclip, '-in', '-selection', self.param.selection)
        rc, _, stderr = run(command, stdin=value)
        if rc:
            fmt = '%s failed with stderr: %s'
            raise ClipboardError(fmt % (self.xclip, stderr))


clipboard_drivers.register(Pasteboard)
clipboard_drivers.register(Xclip)



# -*- coding: utf-8 -*-
"""
Secret generator API.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
import random
import string

from clik.util import AttributeDict

#: "Registry" for secret generators. Keys are the "friendly" name for
#: the generator and the values are the generator functions.
#:
#: :type: :class:`dict` mapping ``str -> fn(int, str)``
generate = AttributeDict()


def generator(name, default=False):
    """
    Register a generator.

    :param str name: "Friendly" name for the generator
    :param bool default: If ``true`` generator will be the one used when
                         ``generate.default`` is called (defaults to
                         ``False``)
    :return: Decorator that returns the function passed in
    """
    def decorator(fn):
        """Register function (possibly as the default) and return it as-is."""
        generate[name] = fn
        if default:
            generate.default = fn
        return fn
    return decorator


class UnsurmountableConstraints(Exception):
    """Raised when a secret cannot be generated given the constraints."""


@generator('random', default=True)
def random_characters(length, disallowed_characters=''):
    """
    Return secret composed of random printable characters.

    Registered as the ``random`` generator, and also as the ``default``.

    :param int length: Length of secret to generate
    :param str disallowed_chars: Characters that may not be present in the
                                 secret
    :return: Randomly-generated string
    :rtype: :class:`str`
    """
    choice = random.SystemRandom().choice
    characters = string.digits \
                 + string.ascii_lowercase \
                 + string.ascii_uppercase \
                 + string.punctuation  # noqa: E127
    for char in disallowed_characters:
        characters = characters.replace(char, '')
    if len(characters) < 1:
        raise UnsurmountableConstraints('no characters to choose from')
    return ''.join([choice(characters) for _ in range(length)])



# -*- coding: utf-8 -*-
"""
Entry point and top-level package information.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
import sys


#: Version of the program.
#:
#: :type: :class:`str`
__version__ = '0.6.0'


def main(argv=None, exit=sys.exit):
    """Entry point for the program."""
    import safe.app
    import safe.cmd
    return safe.app.safe.main(argv, exit)



# -*- coding: utf-8 -*-
"""
New policy command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
from __future__ import print_function

import sys

from clik import args, g, parser

from safe.cmd.new import new
from safe.ec import VALIDATION_ERROR
from safe.form.policy import NewPolicyForm


@new(alias='p')
def policy():
    """Add a policy to the database."""
    form = NewPolicyForm()
    form.configure_parser(exclude=['name'])

    parser.add_argument(
        'name',
        nargs=1,
        help='name for the new policy',
    )

    yield

    args.name = args.name[0]
    if not form.bind_and_validate():
        msg = 'error: there were validation error(s) with input value(s)'
        print(msg, file=sys.stderr)
        form.print_errors()
        yield VALIDATION_ERROR
    form.create_policy()
    g.commit_and_save()



# -*- coding: utf-8 -*-
"""
Root of the :mod:`clik` application.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
from __future__ import print_function

import getpass
import os
import sys

from clik import app, args, g, parser, run_children

from safe import __version__
from safe.db import open_database
from safe.ec import CANCELED, DECRYPTION_FAILED, FILE_ARGUMENT_REQUIRED, \
    MISSING_FILE, MISSING_GPG, SRM_FAILED
from safe.gpg import GPGError, GPGFile, PREFERRED_CIPHER
from safe.model import orm
from safe.srm import secure_delete, SecureDeleteError
from safe.util import expand_path, prompt_bool, temporary_directory


#: Argument name for the internal value that indicates whether the
#: file argument is required.
#:
#: :type: :class:`str`
IGNORE_FILE_ARGUMENT = '_safe_ignore_file_argument'


def ignore_file_argument(ignore=True):
    """
    Configure whether safe ignores the file argument for a command.

    Example::

        @safe
        def gen():
            # "Generate ad-hoc secret" command does not need a file
            ignore_file_argument()
            yield
            # Do stuff

        @gen
        def per_policy():
            # ...but gen's child command per-policy does need the file
            ignore_file_argument(False)
            yield
            # Do stuff

    """
    parser.set_defaults(**{IGNORE_FILE_ARGUMENT: ignore})


@app
def safe():
    """
    Password manager for people who like GPG and the command line.

    For more information, see the full project documentation at
    https://decafjoe-safe.readthedocs.io.
    """
    parser.add_argument(
        '--version',
        action='version',
        version='%%(prog)s %s' % __version__,
    )
    parser.add_argument(
        '-f',
        '--file',
        help='path to gpg-encrypted sqlite database',
    )
    parser.add_argument(
        '-c',
        '--cipher',
        default=PREFERRED_CIPHER,
        help='cipher to use for encryption (default: %(default)s)',
        metavar='CIPHER',
    )

    yield

    if getattr(args, IGNORE_FILE_ARGUMENT, False):
        yield run_children()

    if args.file is None:
        msg = 'error: -f/--file argument is required with this command'
        print(msg, file=sys.stderr)
        yield FILE_ARGUMENT_REQUIRED

    path = expand_path(args.file)
    if not os.path.exists(path):
        print('error: database file does not exist:', path, file=sys.stderr)
        yield MISSING_FILE

    def print_error(message, stdout, stderr, path_=None):
        if path_ is None:
            path_ = path
        print('error: %s: %s' % (message, path_), file=sys.stderr)
        if stdout:
            print('\nstdout:\n%s' % stdout, file=sys.stderr)
        if stderr:
            print('\nstderr:\n%s' % stderr, file=sys.stderr)

    try:
        gpg_file = GPGFile(path)
    except GPGError as e:
        print_error(e.message, e.stdout, e.stderr)
        yield MISSING_GPG

    password = None
    if gpg_file.symmetric:
        password = getpass.getpass()

    try:
        with temporary_directory() as tmp:
            plaintext_path = os.path.join(tmp, 'db')

            while True:
                try:
                    gpg_file.decrypt_to(plaintext_path, password)
                    break
                except GPGError as e:
                    print_error(e.message, e.stdout, e.stderr)
                    print(file=sys.stderr)
                    prompt = 'Command failed. Try again?'
                    if prompt_bool(prompt, default=False):
                        print('\n\n', file=sys.stderr)
                        if gpg_file.symmetric:
                            password = getpass.getpass()
                    else:
                        yield DECRYPTION_FAILED

            try:
                g.db = open_database(plaintext_path)
                cipher = args.cipher

                def commit_and_save():
                    """
                    Commit outstanding db changes and save encrypted file.

                    :raise: :exc:`safe.gpg.GPGError` if encryption fails
                    """
                    g.db.commit()
                    gpg_file.save(plaintext_path, cipher=cipher)

                g.commit_and_save = commit_and_save
                with orm.bind(g.db):
                    ec = run_children()
                    if ec:
                        yield ec
            finally:
                try:
                    secure_delete(plaintext_path)
                except SecureDeleteError as e:
                    print_error(e.message, e.stdout, e.stderr)
                    yield SRM_FAILED
    except KeyboardInterrupt:
        print('canceled by user', file=sys.stderr)
        yield CANCELED



# -*- coding: utf-8 -*-
"""
No project is complete without a utility module.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
import contextlib
import os
import shutil
import stat
import subprocess
import tempfile

from safe.compat import input


def expand_path(path):
    """
    Return absolute path, with variables and ``~`` expanded.

    :param str path: Path, possibly with variables and ``~``
    :return: Absolute path with special sequences expanded
    :rtype: str
    """
    return os.path.abspath(os.path.expanduser(os.path.expandvars(path)))


def get_executable(name):
    """
    Return the full path to executable named ``name``, if it exists.

    :param str name: Name of the executable to find
    :return: Full path to the executable or ``None``
    :rtype: :class:`str` or ``None``
    """
    directories = filter(None, os.environ.get('PATH', '').split(os.pathsep))
    for directory in directories:
        path = os.path.join(directory.strip('"'), name)
        if os.path.isfile(path) and os.access(path, os.X_OK):
            return path


def prompt_bool(prompt, default=False):
    """
    Prompt user for a yes or no answer and return the result as a boolean.

    :param str prompt: Prompt. ``' [y/n] '`` will be appended to this value
    :param bool default: Default value if user enters nothing
    :return: Boolean indicating user's choice
    :rtype: :class:`bool`
    """
    if default is True:
        choices = ' [Y/n] '
        flip_if = 'n'
    else:
        choices = ' [y/N] '
        flip_if = 'y'
    response = input(prompt + choices)
    if response and response[0].lower() == flip_if:
        return not default
    return default


@contextlib.contextmanager
def temporary_directory():
    """
    Context manager that creates a temporary directory for use in the body.

    Example::

       with temporary_directory() as tmp:
           # do stuff with tmp

    The temporary directory permissions are set to 0700 before handing control
    over to the body.
    """
    tmp = tempfile.mkdtemp()
    os.chmod(tmp, stat.S_IRWXU)
    try:
        yield tmp
    finally:
        shutil.rmtree(tmp)


class Subprocess(subprocess.Popen):
    """Subclass whose :meth:`communicate` method turns bytes into strings."""

    def communicate(self, stdin=None):
        """
        Override parent to make sure bytes are decoded into strings.

        :param stdin: Data to send to stdin
        :type stdin: :class:`str` or ``None``
        :return: 2-tuple, ``(stdout, stderr)``
        :rtype: 2-:func:`tuple`
        """
        if stdin is not None:
            stdin = stdin.encode('utf-8')
        stdout, stderr = super(Subprocess, self).communicate(stdin)
        if stdout is not None:
            stdout = stdout.decode('utf-8')
        if stderr is not None:
            stderr = stderr.decode('utf-8')
        return stdout, stderr



# -*- coding: utf-8 -*-
"""
Secret generation for stored policies.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
from __future__ import print_function

import sys

from clik import args, parser

from safe.app import ignore_file_argument
from safe.cmd.gen import gen
from safe.ec import NO_SUCH_POLICY, VALIDATION_ERROR
from safe.model import Policy


DEFAULT_COUNT = 1


@gen(name='per-policy', alias='pp')
def per_policy():
    """Generate random strings to stdout using policy from database."""
    ignore_file_argument(False)

    parser.add_argument(
        'policy',
        help='name of policy',
        nargs=1,
    )
    parser.add_argument(
        '-c',
        '--count',
        default=DEFAULT_COUNT,
        help='number of secrets to generate (one per line) (default: '
             '%(default)s)',
        type=int,
    )

    yield

    if args.count < 1:
        print('error: -c/--count must be one or greater', file=sys.stderr)
        yield VALIDATION_ERROR

    policy = Policy.for_name(args.policy[0])
    if policy is None:
        print('error: no policy named', args.policy[0], file=sys.stderr)
        yield NO_SUCH_POLICY

    for _ in range(args.count):
        print(policy.generate_secret())



# -*- coding: utf-8 -*-
"""
New account command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
from __future__ import print_function

import getpass
import sys

from clik import args, g, parser

from safe.cmd.new import new
from safe.ec import VALIDATION_ERROR
from safe.form.account import NewAccountForm
from safe.model import Password


@new(alias='a')
def account():
    """Add an account to the database."""
    form = NewAccountForm()
    form.configure_parser(exclude=['name'])

    parser.add_argument(
        'name',
        nargs=1,
        help='name for the new account',
    )
    parser.add_argument(
        '-p',
        '--password',
        action='store_true',
        default=False,
        help='set the password for the new account (prompts for value)',
    )

    yield

    args.name = args.name[0]
    if not form.bind_and_validate():
        msg = 'error: there were validation error(s) with input value(s)'
        print(msg, file=sys.stderr)
        form.print_errors()
        yield VALIDATION_ERROR

    account = form.create_account()

    if args.password:
        while True:
            password = getpass.getpass('Password: ')
            confirm = getpass.getpass('Confirm: ')
            if password == confirm:
                break
            print('error: passwords did not match\n', file=sys.stderr)
        g.db.commit()
        g.db.add(Password(account_id=account.id, value=password))

    g.commit_and_save()



# -*- coding: utf-8 -*-
"""
Drop policy command.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
from __future__ import print_function

import sys

from clik import args, g, parser

from safe.cmd.drop import drop
from safe.ec import CANCELED, NO_SUCH_POLICY
from safe.model import Account, Policy
from safe.util import prompt_bool


@drop(alias='p')
def policy():
    """Drop a policy from the database."""
    parser.add_argument(
        'name',
        nargs=1,
        help='name of policy to drop',
    )

    yield

    policy = Policy.for_name(args.name[0])
    if policy is None:
        print('error: no policy named', args.name[0], file=sys.stderr)
        yield NO_SUCH_POLICY

    accounts = {}
    related = (('password', 'passwords'), ('question', 'security questions'))
    for attr, description in related:
        query = Account.query\
                       .filter_by(**{'%s_policy_id' % attr: policy.id})\
                       .all()
        for account in query:
            if account.name in accounts:
                accounts[account.name] += ', %s' % description
            else:
                accounts[account.name] = description

    if len(accounts) > 0:
        print()
        fmt = 'The policy "%s" is associated with the following accounts.'
        print(fmt % policy.name)
        msg = 'The accounts will be disassociated with the policy before it ' \
              'is deleted, but will otherwise remain intact.'
        print(msg)
        fmt = '- %s (%s)'
        for name in sorted(accounts):
            print(fmt % (name, accounts[name]))
        print()

    msg = 'Delete policy "%s"?' % policy.name
    if not prompt_bool(msg, default=False):
        yield CANCELED

    g.db.delete(policy)
    g.commit_and_save()



# -*- coding: utf-8 -*-
"""
Python compatibility helpers.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
try:
    input = raw_input
except NameError:
    try:
        input = __builtins__['input']
    except TypeError:
        input = __builtins__.input



# -*- coding: utf-8 -*-
"""
Facilities for interacting with GPG encrypted files.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
import os
import re
import shutil
import subprocess

from safe.util import get_executable, Subprocess, temporary_directory


#: Name of the cipher to use if unspecified in :meth:`GPGFile.save`.
#:
#: :type: :class:`str`
PREFERRED_CIPHER = 'aes256'


class GPGError(Exception):
    """Raised for errors from this module."""

    def __init__(self, message, stdout, stderr):
        """
        Instantiate the error.

        :param str message: Short message describing the error
        :param stdout: Standard output related to the error
        :type stdout: :class:`str` or ``None``
        :param stderr: Standard error related to the error
        :type stderr: :class:`str` or ``None``
        """
        super(GPGError, self).__init__(message)

        #: Short message describing the error.
        #:
        #: :type: :class:`str`
        self.message = message

        #: Standard out associated with the error.
        #:
        #: :type: :class:`str` or ``None``
        self.stdout = stdout

        #: Standard error associated with the error.
        #:
        #: :type: :class:`str` or ``None``
        self.stderr = stderr


def get_gpg_executable():
    """
    Return GPG executable, raising a :exc:`GPGError` if not found.

    This will first look for an executable named ``gpg2``, returning it
    immediately if found. If ``gpg2`` does not exist but ``gpg`` does, this
    function runs ``gpg --version`` to check the version. If version 2, the
    absolute path to the executable is returned.

    Failure to find a GPG2 executable results in a :exc:`GPGError` being
    raised.

    :raise: :exc:`GPGError` if GPG executable is not found
    :return: Absolute path to the GPG executable
    :rtype: :class:`str`
    """
    rv = get_executable('gpg2')
    if rv is not None:
        return rv

    rv = get_executable('gpg')
    if rv is None:
        msg = 'neither gpg2 nor gpg executables were found'
        raise GPGError(msg, None, None)

    process = Subprocess((rv, '--version'), stdout=subprocess.PIPE)
    stdout, stderr = process.communicate()
    if process.returncode:
        msg = '`gpg --version` exited non-zero: %s' % process.returncode
        raise GPGError(msg, stdout, stderr)

    regex = re.compile(r'^gpg[^\n]+\s+((?P<major>\d+)\.\d+\.\d+).+', re.DOTALL)
    match = regex.search(stdout)
    if not match:
        msg = 'could not extract version from `gpg --version`'
        raise GPGError(msg, stdout, stderr)

    major_version = match.groupdict()['major']
    if major_version != '2':
        msg = 'safe requires gpg version 2, found version: %s' % major_version
        raise GPGError(msg, stdout, stderr)

    return rv


class GPGSubprocess(Subprocess):
    """Convenience class for running GPG commands."""

    #: Absolute path to the GPG executable.
    #:
    #: This is populated the first time a :class:`GPGSubprocess` is
    #: instantiated and is subsequently reused.
    #:
    #: :type: ``None`` until populated with a :class:`str` on :meth:`__init__`
    _gpg = None

    def __init__(self, command):
        """
        Instantiate the subprocess.

        :param command: Arguments to pass to GPG
        :type command: :func:`tuple` of arguments, *not* including ``gpg``
                       itself at the beginning
        """
        if self.__class__._gpg is None:
            self.__class__._gpg = get_gpg_executable()

        cmd = (self._gpg,) + command
        pipe = subprocess.PIPE
        kwargs = dict(stdin=pipe, stdout=pipe, stderr=pipe)
        super(GPGSubprocess, self).__init__(cmd, **kwargs)


class GPGFile(object):
    """Manage decryption and encryption of a GPG file."""

    #: Regex matching the keyid output string from ``gpg
    #: --list-packets``.
    #:
    #: :type: :func:`re.compile`
    KEYID_RE = re.compile(r'keyid (?P<keyid>[0-9A-F]+)')

    # These are defined here so we can reference them in the internal
    # documentation in doc/development/internals/gpg.rst.
    _homedir = None
    _keyid = None
    _password = None
    _path = None
    _symmetric = None

    def __init__(self, path):
        """
        Instantiate the file wrapper.

        :param str path: Path to the GPG encrypted file
        :raise: :exc:`GPGError` if file cannot be read
        """
        #: Home directory to use for GnuPG calls (i.e. the ``--homedir``
        #: argument). Defaults to ~/.gnupg. This attribute exists to
        #: allow tests to tweak the GnuPG environment while running,
        #: and is not otherwise used.
        #:
        #: :type: :class:`str`
        self._homedir = os.path.join(os.path.expanduser('~'), '.gnupg')

        #: Keyid to which the file was encrypted. Populated when
        #: :meth:`decrypt_to` is called.
        #:
        #: :type: :class:`str` or ``None`` (if file is symmetrically
        #:        encrypted)
        self._keyid = None

        #: Password with which file was encryted. Populated when
        #: :meth:`decrypt_to` is called.
        #:
        #: :type: :class:`str` or ``None`` (if file is asymmetrically
        #:        encrypted)
        self._password = None

        #: Path to the encrypted file.
        #:
        #: :type: :class:`str`
        self._path = path

        #: Boolean indicating whether the file is symmetrically
        #: encrypted. If false, the file is asymmetrically encrypted.
        #:
        #: :type: :class:`bool`
        self._symmetric = None

        with temporary_directory() as tmp:
            command = (
                '--batch',
                '--homedir', tmp,
                '--passphrase', '',
                '--quiet',
                '--list-packets',
                path,
            )
            process = GPGSubprocess(command)
            stdout, stderr = process.communicate()

        for line in stdout.splitlines():
            if line.startswith(':symkey'):
                self._symmetric = True
                break
            elif line.startswith(':pubkey'):
                self._symmetric = False
                error_msg = 'failed to extract keyid from packets'
                match = self.KEYID_RE.search(line)
                if not match:
                    raise GPGError(error_msg, stdout, stderr)
                keyid = match.groupdict()['keyid']
                if re.search('^0+$', keyid):
                    raise GPGError(error_msg, stdout, stderr)
                self._keyid = keyid
                break

        if self._symmetric is None:
            msg = 'did not find encryption type packet in file (are you ' \
                  'sure this is a gpg file?)'
            raise GPGError(msg, stdout, stderr)

    @property
    def symmetric(self):
        """If true, file is encrypted symmetrically (i.e. with a password)."""
        return self._symmetric

    def decrypt_to(self, path, password=None):
        """
        Decrypt file to ``path`` using ``password``.

        If decryption is successful, this will cache the password/keyid for
        use in subsequent calls to :meth:`save`.

        :param str path: Path to which to decrypt file
        :param password: Password for file, if encrypted symmetrically
        :type password: :class:`str` if file is symmetrically encrypted else
                        ``None``
        :raise: :exc:`GPGError` if decryption fails
        :rtype: ``None``
        """
        if self.symmetric and password is None:
            raise Exception('password required when symmetrically encrypted')
        command = (
            '--batch',
            '--homedir', self._homedir,
            '--output', path,
            '--quiet',
        )
        if self.symmetric:
            command += ('--passphrase-fd', '0')
        command += ('--decrypt', self._path)
        process = GPGSubprocess(command)
        stdout, stderr = process.communicate(password)
        if process.returncode:
            raise GPGError('failed to decrypt file', stdout, stderr)
        self._password = password

    def save(self, source, cipher=PREFERRED_CIPHER):
        """
        Save plaintext file ``source`` back to the original path, encrypted.

        :meth:`decrypt_to` **must be called before calling this method.**
        Certain values needed by this method are cached when a file is
        decrypted. (Namely, password for symmetrically encrypted files and
        keyid for asymmetrically encrypted files.)

        :param str source: Path to file to save
        :param str cipher: Cipher to use for encryption (defaults to
                           :data:`PREFERRED_CIPHER`)
        :raise: :exc:`GPGError` if encryption fails (original encrypted file is
                left untouched)
        :rtype: ``None``
        """
        with temporary_directory() as tmp:
            tmp_path = os.path.join(tmp, 'f')
            command = (
                '--armor',
                '--batch',
                '--cipher-algo', cipher,
                '--homedir', self._homedir,
                '--output', tmp_path,
                '--quiet',
            )
            if self.symmetric:
                command += ('--passphrase-fd', '0', '--symmetric')
            else:
                command += ('--recipient', self._keyid, '--encrypt')
            command += (source,)
            process = GPGSubprocess(command)
            stdout, stderr = process.communicate(self._password)
            if process.returncode:
                raise GPGError('failed to re-encrypt file', stdout, stderr)
            shutil.move(tmp_path, self._path)



# -*- coding: utf-8 -*-
"""
Account form definitions.

:author: Joe Joyce <joe@decafjoe.com>
:copyright: Copyright (c) Joe Joyce and contributors, 2016-2017.
:license: BSD
"""
from clik import g
from clik_wtforms import FieldList, Form, StringField
from wtforms.validators import InputRequired, ValidationError

from safe.form import slug_validator
from safe.model import Account, Alias, Code, Policy, Question


def policy_validator(_, field):
    """
    Validate that ``field.data`` is the name of a policy.

    :param field: Field to validate
    :type field: :class:`wtforms.fields.core.Field`
    :raises: :exc:`wtforms.validators.ValidationError` if data does not
             contain the name of a policy
    """
    if field.data and Policy.id_for_name(field.data) is None:
        raise ValidationError('No policy with that name')


class Operation(object):
    """Enum of "operation" types for certain form fields."""

    #: Update answer to a security question.
    A = 'a'

    #: Add a new item.
    ADD = ''

    #: Create a new security question.
    NEW = 'new'

    #: Update question part of a security question.
    Q = 'q'

    #: Remove an item.
    REMOVE = 'rm'

    #: Mark a backup code as used.
    USED = 'used'


class AccountForm(Form):
    """Base form containing common fields, not meant to be used directly."""

    #: Short one-line description for the account.
    #:
    #: :type: :class:`clik_wtforms.StringField`
    description = StringField(
        description='short one-line description for account',
    )

    #: Email associated with the account.
    #:
    #: :type: :class:`clik_wtforms.StringField`
    email = StringField(description='email associated with account')

    #: Name of the question policy for the account.
    #:
    #: :type: :class:`clik_wtforms.StringField` validated by
    #:        :func:`policy_validator`
    question_policy = StringField(
        description='policy to apply to security question answers for the '
                    'account',
        metavar='POLICY',
        validators=[policy_validator],
    )

    #: Name of the password policy for the account.
    #:
    #:
    #: :type: :class:`clik_wtforms.StringField` validated by
    #:        :func:`policy_validator`
    password_policy = StringField(
        description='policy to apply to passwords for the account',
        metavar='POLICY',
        validators=[policy_validator],
    )

    #: Username associated with the account.
    #:
    #: :type: :class:`clik_wtforms.StringField`
    username = StringField(description='username associated with the account')

    @staticmethod
    def get_short_arguments():
        """Return short arguments for the base fields."""
        return dict(d='description', e='email', u='username')

    def update_account(self, account):
        """
        Modify ``account`` based on the values from the form.

        :param safe.model.Account account: Account to update
        """
        if self.description.data is not None:
            account.description = self.description.data
        if self.email.data is not None:
            account.email = self.email.data
        if self.question_policy.data is not None:
            policy_id = Policy.id_for_name(self.question_policy.data)
            account.question_policy_id = policy_id
        if self.password_policy.data is not None:
            policy_id = Policy.id_for_name(self.password_policy.data)
            account.password_policy_id = policy_id
        if self.username.data is not None:
            account.username = self.username.data


class NewAccountForm(AccountForm):
    """Form for creating new accounts."""

    #: Alias(es) associated with the account.
    #:
    #: :type: :class:`clik_wtforms.FieldList` of
    #:        :class:`clik_wtforms.StringField` validated by
    #:        :func:`safe.form.slug_validator`
    alias = FieldList(
        StringField(validators=[slug_validator]),
        description='alias for the account',
    )

    #: Backup code(s) associated with the account.
    #:
    #: :type: :class:`clik_wtforms.FieldList` of
    #:        :class:`clik_wtforms.StringField`
    code = FieldList(
        StringField(),
        description='backup code for the account',
    )

    #: Name for the account.
    #:
    #: :type: :class:`clik_wtforms.StringField` validated by
    #:        :func:`safe.form.slug_validator`
    name = StringField(
        description='name for the account',
        validators=[InputRequired(), slug_validator],
    )

    @classmethod
    def get_short_arguments(cls):
        """Return short arguments, merged with parent form short args."""
        d = super(NewAccountForm, cls).get_short_arguments()
        d.update(dict(a='alias', c='code', n='name'))
        return d

    def validate_alias(self, field):
        """
        Validate that aliases are unique.

        :raises: :exc:`wtforms.validators.ValidationError` if alias is
                 supplied twice or is the same as an existing account name
                 or alias
        """
        names = [self.name.data]
        for alias in field.data:
            if alias in names:
                fmt = 'Alias "%s" already supplied as name or other alias'
                raise ValidationError(fmt % alias)
            if Account.id_for_slug(alias):
                fmt = 'Account with name/alias "%s" already exists'
                raise ValidationError(fmt % alias)
            names.append(alias)

    def validate_name(self, field):
        """
        Validate that the new name does not already exist.

        :raises: :exc:`wtforms.validators.ValidationError` if new name is
                 the same as the name or alias of an existing account
        """
        if Account.id_for_slug(field.data):
            msg = 'Account with that name/alias already exists'
            raise ValidationError(msg)

    def create_account(self):
        """
        Create a new account based on the form data.

        Note that the new account object will be commited to ``g.db`` in
        order to get its id, which is required to create any new aliases
        or code instances in the database.

        :return: Newly-created account
        :rtype: :class:`safe.model.Account`
        """
        account = Account(name=self.name.data)
        super(NewAccountForm, self).update_account(account)
        g.db.add(account)
        g.db.commit()
        g.db.refresh(account)
        for alias in self.alias.data:
            g.db.add(Alias(account_id=account.id, value=alias))
        for code in self.code.data:
            g.db.add(Code(account_id=account.id, value=code))
        return account


class UpdateAccountForm(AccountForm):
    """Form for updating an existing account."""

    #: Alias(es) associated with the account. This is also used to
    #: remove aliases, by passing the value ``"rm:alias-name"``.
    #:
    #: :type: :class:`clik_wtforms.FieldList` of
    #:        :class:`clik_wtforms.StringField`
    alias = FieldList(
        StringField(),
        description='add or remove alias for the account',
    )

    #: Backup code(s) associated with the account. This can also be
    #: used to remove codes, by passing ``"rm:code"``, or mark a code
    #: as used, by passing ``"used:code"``.
    #:
    #: :type: :class:`clik_wtforms.FieldList` of
    #:        :class:`clik_wtforms.StringField`
    code = FieldList(
        StringField(),
        description='add, remove, or "mark as used" a backup code for the '
                    'account',
    )

    #: New name for the account.
    #:
    #: :type: :class:`clik_wtforms.StringField`
    new_name = StringField(
        description='new name for this account (replaces current name)',
        metavar='NAME',
        validators=[slug_validator],
    )

    #: Security question(s) associated with the account. This field
    #: supports the following "operations:"
    #:
    #: * Add new security question by passing ``"new:identifier"``
    #: * Remove question by passing ``"rm:identifier"``
    #: * Update question text by passing ``"q:identifier:new-text"``
    #: * Update answer text by passing ``"a:identifier:new-text"``
    #:
    #: :type: :class:`clik_wtforms.FieldList` of
    #:        :class:`clik_wtforms.StringField`
    question = FieldList(
        StringField(),
        description='add, remove, or update security questions/answers '
                    'associated with the account',
    )

    @classmethod
    def get_short_arguments(cls):
        """Return short arguments, merged with parent form short args."""
        d = super(UpdateAccountForm, cls).get_short_arguments()
        d.update(dict(a='alias', c='code', n='new_name', q='question'))
        return d

    def bind_and_validate(self, account, args=None):
        """Bind the ``account`` to the form, then call the superclass."""
        self.account = account
        return super(UpdateAccountForm, self).bind_and_validate(args)

    def validate_alias(self, field):
        """
        Validate alias operations.

        :param clik_wtforms.FieldList field: Alias field
        :raises: :exc:`wtforms.validators.ValidationError` if any operation
                 is invalid, or refers to a non-existant alias
        """
        field.operations = []
        for value in field.data:
            if ':' in value:
                op, subject = value.split(':', 1)
                if op == Operation.REMOVE:
                    obj = self.account.alias_query\
                                      .filter_by(value=subject)\
                                      .first()
                    if obj is None:
                        fmt = 'No alias named "%s" associated with account'
                        raise ValidationError(fmt % subject)
                    if [op, obj] in field.operations:
                        fmt = 'Alias "%s" already scheduled for removal'
                        raise ValidationError(fmt % subject)
                    field.operations.append([op, obj])
                else:
                    raise ValidationError('Unknown operation "%s"' % op)
            else:
                op = Operation.ADD
                if Account.id_for_slug(value):
                    fmt = 'Account with name/alias "%s" already exists'
                    raise ValidationError(fmt % value)
                if [op, value] in field.operations:
                    fmt = 'Alias "%s" already scheduled for addition'
                    raise ValidationError(fmt % value)
                field.operations.append([op, value])

    def validate_code(self, field):
        """
        Validate code operations.

        :param clik_wtforms.FieldList field: Code field
        :raises: :exc:`wtforms.validators.ValidationError` if any operation
                 is invalid, or refers to a non-existant code
        """
        field.operations = []
        for value in field.data:
            if ':' in value:
                op, subject = value.split(':', 1)
                if op in (Operation.REMOVE, Operation.USED):
                    obj = self.account.code_query\
                                      .filter_by(value=subject)\
                                      .first()
                    if obj is None:
                        fmt = 'No code "%s" associated with account'
                        raise ValidationError(fmt % subject)
                    if [Operation.REMOVE, obj] in field.operations:
                        fmt = 'Code "%s" already scheduled for removal'
                        raise ValidationError(fmt % subject)
                    if [Operation.USED, obj] in field.operations:
                        fmt = 'Code "%s" already scheduled to be marked used'
                        raise ValidationError(fmt % subject)
                    field.operations.append([op, obj])
                else:
                    raise ValidationError('Unknown operation "%s"' % op)
            else:
                op = Operation.ADD
                code = self.account.code_query.filter_by(value=value).first()
                if code is not None:
                    fmt = 'Code "%s" is already associated with this account'
                    raise ValidationError(fmt % value)
                if [op, value] in field.operations:
                    fmt = 'Code "%s" already scheduled for addition'
                    raise ValidationError(fmt % value)
                field.operations.append([op, value])

    def validate_new_name(self, field):
        """
        Validate new name.

        :param clik_wtforms.StringField field: New name field
        :raises: :exc:`wtforms.validators.ValidationError` if name is the same
                 as the current name, or already exists as a name or alias
                 of another account
        """
        field.change_name = False
        if field.data:
            if field.data == self.account.name:
                msg = 'New name is the same as the current name'
                raise ValidationError(msg)
            if Account.id_for_slug(field.data) is not None:
                fmt = 'Account with name/alias "%s" already exists'
                raise ValidationError(fmt % field.data)
            field.change_name = True

    def validate_question(self, field):
        """
        Validate question operations.

        :param clik_wtforms.FieldList field: Question field
        :raises: :exc:`wtforms.validators.ValidationError` if any operation
                 is invalid, or refers to a non-existant question identifier
        """
        field.operations = []
        for value in field.data:
            if ':' in value:
                op, subject = value.split(':', 1)
                if op in (Operation.A, Operation.Q, Operation.REMOVE):
                    details = None
                    if op in (Operation.A, Operation.Q):
                        if ':' in subject:
                            subject, details = subject.split(':', 1)
                        for other_op, other_subject, _ in field.operations:
                            if isinstance(other_subject, Question):
                                other_subject = other_subject.identifier
                            if subject == other_subject:
                                if op == other_op:
                                    fmt = 'Redundant "%s" operation for ' \
                                          'question with identifier "%s"'
                                    raise ValidationError(fmt % (op, subject))
                                if other_op == Operation.REMOVE:
                                    fmt = 'Question with identifier "%s" ' \
                                          'already scheduled for removal'
                                    raise ValidationError(fmt % subject)
                    else:
                        for other_op, other_subject, _ in field.operations:
                            if isinstance(other_subject, Question):
                                other_subject = other_subject.identifier
                            if subject == other_subject:
                                if op == other_op:
                                    fmt = 'Question with identifier "%s" ' \
                                          'already scheduled for removal'
                                    raise ValidationError(fmt % subject)
                                else:
                                    fmt = 'Question with identifier "%s" ' \
                                          'already scheduled to be updated'
                                    raise ValidationError(fmt % subject)
                    obj = self.account.question_query\
                                      .filter_by(identifier=subject)\
                                      .first()
                    if obj is None:
                        fmt = 'No question with identifier "%s" associated ' \
                              'with account'
                        raise ValidationError(fmt % subject)
                    field.operations.append([op, obj, details])
                elif op == Operation.NEW:
                    question = self.account.question_query\
                                           .filter_by(identifier=subject)\
                                           .first()
                    if question is not None:
                        fmt = 'Question with identifier "%s" is already ' \
                              'associated with this account'
                        raise ValidationError(fmt % subject)
                    if [op, subject, None] in field.operations:
                        fmt = 'Question with identifier "%s" already ' \
                              'scheduled for addition'
                        raise ValidationError(fmt % subject)
                    field.operations.append([op, subject, None])
                else:
                    raise ValidationError('Unknown operation "%s"' % op)
            else:
                raise ValidationError('No operation specified')

    def update_account(self):
        """
        Update the bound account based on the data in the form.

        Note that this does not commit the changes.
        """
        super(UpdateAccountForm, self).update_account(self.account)

        if self.new_name.change_name:
            self.account.name = self.new_name.data

        for op, subject in self.alias.operations:
            if op == Operation.ADD:
                g.db.add(Alias(account_id=self.account.id, value=subject))
            elif op == Operation.REMOVE:
                g.db.delete(subject)
            else:  # pragma: no cover (unreachable)
                raise Exception('unreachable')

        for op, subject in self.code.operations:
            if op == Operation.ADD:
                g.db.add(Code(account_id=self.account.id, value=subject))
            elif op == Operation.REMOVE:
                g.db.delete(subject)
            elif op == Operation.USED:
                subject.used = True
            else:  # pragma: no cover (unreachable)
                raise Exception('unreachable')

        for op, subject, details in self.question.operations:
            if op == Operation.A:
                subject.answer = ''
                if details:
                    subject.answer = details
            elif op == Operation.NEW:
                question, answer = '', ''
                if details:
                    question, answer = details
                g.db.add(Question(
                    account_id=self.account.id,
                    answer=answer,
                    identifier=subject,
                    question=question,
                ))
            elif op == Operation.Q:
                subject.question = ''
                if details:
                    subject.question = details
            elif op == Operation.REMOVE:
                g.db.delete(subject)
            else:  # pragma: no cover (unreachable)
                raise Exception('unreachable')
